@@ -6,21 +6,21 @@ API module
 Provides high-level interface for programmatic API interactions
 """
 # Builtins
-import json
 import logging
+from datetime import datetime
 from typing import Optional
-from datetime import datetime, timedelta
 from urllib import parse
 
+from conjur_api.errors.errors import HttpStatusError, InvalidResourceException, MissingRequiredParameterException
 # Internals
+from conjur_api.errors.errors import MissingApiTokenException
 from conjur_api.http.endpoints import ConjurEndpoint
-from conjur_api.interface.credentials_store_interface import CredentialsProviderInterface
+from conjur_api.interface.authentication_strategy_interface import AuthenticationStrategyInterface
+# pylint: disable=too-many-instance-attributes
+from conjur_api.models import Resource, ConjurConnectionInfo, ListPermittedRolesData, \
+    ListMembersOfData, CreateHostData, CreateTokenData, SslVerificationMetadata, SslVerificationMode
 from conjur_api.wrappers.http_response import HttpResponse
 from conjur_api.wrappers.http_wrapper import HttpVerb, invoke_endpoint
-from conjur_api.errors.errors import InvalidResourceException, MissingRequiredParameterException
-# pylint: disable=too-many-instance-attributes
-from conjur_api.models import Resource, ConjurConnectionInfo, ListPermittedRolesData, ListMembersOfData, CreateHostData, \
-    CreateTokenData, SslVerificationMetadata, SslVerificationMode
 
 
 # pylint: disable=unspecified-encoding,too-many-public-methods
@@ -29,11 +29,9 @@ class Api:
     This module provides a high-level programmatic access to the HTTP API
     when all the needed arguments and parameters are well-known
     """
-    # Tokens should only be reused for 5 minutes (max lifetime is 8 minutes)
-    API_TOKEN_DURATION = 5
 
     KIND_VARIABLE = 'variable'
-    KIND_HOSTFACTORY = 'host_factory'
+    KIND_HOST_FACTORY = 'host_factory'
     ID_FORMAT = '{account}:{kind}:{id}'
     ID_RETURN_PREFIX = '{account}:{kind}:'
 
@@ -45,7 +43,7 @@ class Api:
     def __init__(
             self,
             connection_info: ConjurConnectionInfo,
-            credentials_provider: CredentialsProviderInterface,
+            authn_strategy: AuthenticationStrategyInterface,
             ssl_verification_mode: SslVerificationMode = SslVerificationMode.TRUST_STORE,
             debug: bool = False,
             http_debug=False,
@@ -54,13 +52,11 @@ class Api:
         self.ssl_verification_data = SslVerificationMetadata(ssl_verification_mode,
                                                              connection_info.cert_file)
 
-        self._account = connection_info.conjur_account
-        self._url = connection_info.conjur_url
-        self._api_key = None
-        self.credentials_provider: CredentialsProviderInterface = credentials_provider
+        self._connection_info = connection_info
+        self.authn_strategy: AuthenticationStrategyInterface = authn_strategy
         self.debug = debug
         self.http_debug = http_debug
-        self.api_token_expiration = None
+        self.api_token_expiration = datetime.now()
         self._login_id = None
 
         self._default_params = {  # TODO remove, pass to invoke endpoint ConjurConnectionInfo
@@ -73,91 +69,48 @@ class Api:
         # if http_debug: enable_http_logging()
 
     @property
-    def api_key(self) -> str:
-        """
-        Property returns api_key. if no api_key we try the password as sometimes
-        @return: api_key
-        """
-        # TODO do not use password after credentials store is fixed to also store API key
-        return self._api_key or self.password
+    def _account(self) -> str:
+        return self._connection_info.conjur_account
 
     @property
-    # pylint: disable=missing-docstring
+    def _url(self) -> str:
+        return self._connection_info.conjur_url
+
+    @property
+    # pylint: disable=missing-docstring,bare-except
     async def api_token(self) -> str:
         """
         @return: Conjur api_token
         """
         if not self._api_token or datetime.now() > self.api_token_expiration:
             logging.debug("API token missing or expired. Fetching new one...")
-            self.api_token_expiration = datetime.now() + timedelta(minutes=self.API_TOKEN_DURATION)
-            self._api_token = await self.authenticate()
-
+            self._api_token, self.api_token_expiration = await self.authenticate()
             return self._api_token
 
         logging.debug("Using cached API token...")
         return self._api_token
 
-    @property
-    def password(self) -> str:
-        """
-        password as being saved inside credentials_provider
-        @return:
-        """
-        return self.credentials_provider.load(self._url).password
-
-    @property
-    def login_id(self) -> str:
-        """
-        @return: The login_id (username)
-        """
-        if not self._login_id:
-            self._login_id = self.credentials_provider.load(self._url).username
-        return self._login_id
-
     async def login(self) -> str:
         """
         This method uses the basic auth login id (username) and password
-        to retrieve an conjur_api key from the server that can be later used to
+        to retrieve a conjur_api key from the server that can be later used to
         retrieve short-lived conjur_api tokens.
         """
-        logging.debug("Logging in to %s...", self._url)
-        password = self.password
-        if not password:
-            raise MissingRequiredParameterException("password requires when login")
-        response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.LOGIN,
-                                         self._default_params, auth=(self.login_id, password),
-                                         ssl_verification_metadata=self.ssl_verification_data)
-        self._api_key = response.text
-        return self.api_key
+        if hasattr(self.authn_strategy, 'login') and callable(self.authn_strategy.login):
+            return await self.authn_strategy.login(
+                self._connection_info,
+                self.ssl_verification_data
+            )
 
-    async def authenticate(self) -> str:
+    async def authenticate(self) -> tuple[str, datetime]:
         """
-        Authenticate uses the api_key to fetch a short-lived conjur_api token that
-        for a limited time will allow you to interact fully with the Conjur
-        vault.
+        Authenticate based on the chosen strategy to fetch a short-lived conjur_api token that
+        for a limited time will allow you to interact fully with the Conjur vault.
         """
-        if not self.api_key and self.login_id and self.password:
-            # TODO we do this since api_key is not provided. it should be stored like username,
-            # password inside credentials_data
-            await self.login()
-
-        if not self.login_id or not self.api_key:
-            raise MissingRequiredParameterException("Missing parameters in "
-                                                    "authentication invocation")
-
-        params = {
-            'login': self.login_id
-        }
-        params.update(self._default_params)
-
-        logging.debug("Authenticating to %s...", self._url)
-        response = await invoke_endpoint(
-            HttpVerb.POST,
-            ConjurEndpoint.AUTHENTICATE,
-            params,
-            self.api_key,
-            ssl_verification_metadata=self.ssl_verification_data)
-        return response.text
+        return await self.authn_strategy.authenticate(
+            self._connection_info,
+            self.ssl_verification_data
+        )
 
     async def resources_list(self, list_constraints: dict = None) -> dict:
         """
@@ -172,16 +125,20 @@ class Api:
         # Remove 'inspect' from query as it is client-side param that shouldn't get to the server.
         inspect = list_constraints.pop('inspect', None) if list_constraints else None
 
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         if list_constraints is not None:
             response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.RESOURCES,
                                              params,
                                              query=list_constraints,
-                                             api_token=await self.api_token,
+                                             api_token=api_token,
                                              ssl_verification_metadata=self.ssl_verification_data)
         else:
             response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.RESOURCES,
                                              params,
-                                             api_token=await self.api_token,
+                                             api_token=api_token,
                                              ssl_verification_metadata=self.ssl_verification_data)
 
         resources = response.json
@@ -198,6 +155,157 @@ class Api:
         # ?tocpath=Developer%7CREST%C2%A0APIs%7C_____17
         return resources
 
+    async def check_privilege(self, kind: str, resource_id: str, privilege: str, role_id: str = None) -> bool:
+        """
+        This method is used to check for a privilege on a resource.
+        """
+        params = {
+            'account': self._account,
+            'kind': kind,
+            'identifier': resource_id,
+            'privilege': privilege,
+            'role': role_id if role_id else ''
+        }
+        params.update(self._default_params)
+
+        try:
+            response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.PRIVILEGE,
+                                  params,
+                                  api_token=await self.api_token,
+                                  ssl_verification_metadata=self.ssl_verification_data)
+            logging.debug(str(response))
+        except HttpStatusError as err:
+            if err.status == 404:
+                return False
+
+            raise err
+
+        return True
+
+    async def get_resource(self, kind: str, resource_id: str) -> dict:
+        """
+        This method is used to fetch a specific resource.
+        """
+        params = {
+            'account': self._account,
+            'kind': kind,
+            'identifier': resource_id
+        }
+        params.update(self._default_params)
+
+        response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.RESOURCE,
+                                         params,
+                                         api_token=await self.api_token,
+                                         ssl_verification_metadata=self.ssl_verification_data)
+
+        resource = response.json
+
+        # To see the full resources response see
+        # https://docs.conjur.org/Latest/en/Content/Developer/Conjur_API_Show_Resources.htm
+        # ?tocpath=Developer%7CREST%C2%A0APIs%7C_____19
+        return resource
+
+    async def resource_exists(self, kind: str, resource_id: str) -> bool:
+        """
+        This method is used to check whether a specific resource exists.
+        """
+        params = {
+            'account': self._account,
+            'kind': kind,
+            'identifier': resource_id
+        }
+        params.update(self._default_params)
+
+        try:
+            await invoke_endpoint(HttpVerb.HEAD, ConjurEndpoint.RESOURCE,
+                                  params,
+                                  api_token=await self.api_token,
+                                  ssl_verification_metadata=self.ssl_verification_data)
+        except HttpStatusError as err:
+            if err.status == 404:
+                return False
+
+            if err.status == 403:
+                return True
+
+            raise
+
+        return True
+
+    async def get_role(self, kind: str, resource_id: str) -> dict:
+        """
+        This method is used to fetch a specific role.
+        """
+        params = {
+            'account': self._account,
+            'kind': kind,
+            'identifier': resource_id
+        }
+        params.update(self._default_params)
+
+        response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.ROLE,
+                                         params,
+                                         api_token=await self.api_token,
+                                         ssl_verification_metadata=self.ssl_verification_data)
+
+        role = response.json
+
+        # To see the full roles response see
+        # https://docs.conjur.org/Latest/en/Content/Developer/Conjur_API_Show_Role.htm
+        # ?tocpath=Developer%7CREST%C2%A0APIs%7C_____15
+        return role
+
+    async def role_memberships(self, kind: str, resource_id: str, direct: bool) -> dict:
+        """
+        This method is used to fetch the memberships of a role
+        """
+        recursive_param = "memberships" if direct else "all"
+        params = {
+            'account': self._account,
+            'kind': kind,
+            'identifier': resource_id,
+            'membership': recursive_param
+        }
+        params.update(self._default_params)
+
+        response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.ROLES_MEMBERSHIPS,
+                                         params,
+                                         api_token=await self.api_token,
+                                         ssl_verification_metadata=self.ssl_verification_data)
+
+        if direct:
+            memberships = map(lambda membership: membership['role'], response.json)
+            return list(memberships)
+
+        return response.json
+
+    async def role_exists(self, kind: str, resource_id: str) -> bool:
+        """
+        This method is used to check whether a specific role exists.
+        """
+        params = {
+            'account': self._account,
+            'kind': kind,
+            'identifier': resource_id
+        }
+        params.update(self._default_params)
+
+        try:
+            await invoke_endpoint(HttpVerb.HEAD, ConjurEndpoint.ROLE,
+                                  params,
+                                  api_token=await self.api_token,
+                                  ssl_verification_metadata=self.ssl_verification_data)
+        except HttpStatusError as err:
+            if err.status == 404:
+                return False
+
+            if err.status == 403:
+                return True
+
+            raise
+
+        return True
+
     async def get_variable(self, variable_id: str, version: str = None) -> Optional[bytes]:
         """
         This method is used to fetch a secret's (aka "variable") value from
@@ -215,14 +323,18 @@ class Api:
                 'version': version
             }
 
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         # pylint: disable=no-else-return
         if version is not None:
             response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.SECRETS, params,
-                                             api_token=await self.api_token, query=query_params,
+                                             api_token=api_token, query=query_params,
                                              ssl_verification_metadata=self.ssl_verification_data)
         else:
             response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.SECRETS, params,
-                                             api_token=await self.api_token,
+                                             api_token=api_token,
                                              ssl_verification_metadata=self.ssl_verification_data)
         return response.content
 
@@ -242,9 +354,13 @@ class Api:
             'variable_ids': ','.join(full_variable_ids),
         }
 
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.BATCH_SECRETS,
                                          self._default_params,
-                                         api_token=await self.api_token,
+                                         api_token=api_token,
                                          ssl_verification_metadata=self.ssl_verification_data,
                                          query=query_params,
                                          )
@@ -269,7 +385,7 @@ class Api:
             raise MissingRequiredParameterException('create_token_data is empty')
 
         create_token_data.host_factory = self.ID_FORMAT.format(account=self._account,
-                                                               kind=self.KIND_HOSTFACTORY,
+                                                               kind=self.KIND_HOST_FACTORY,
                                                                id=create_token_data.host_factory)
 
         # parse.urlencode, If any values in the query arg are sequences and doseq is true, each
@@ -280,11 +396,16 @@ class Api:
 
         params = {}
         params.update(self._default_params)
+
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         return await invoke_endpoint(HttpVerb.POST,
                                      ConjurEndpoint.HOST_FACTORY_TOKENS,
                                      params,
                                      create_token_data,
-                                     api_token=await self.api_token,
+                                     api_token=api_token,
                                      ssl_verification_metadata=self.ssl_verification_data,
                                      headers={'Content-Type': 'application/x-www-form-urlencoded'})
 
@@ -294,7 +415,7 @@ class Api:
         """
         if create_host_data is None:
             raise MissingRequiredParameterException('create_host_data is empty')
-        request_body_parameters = parse.urlencode(create_host_data.get_host_id())
+        request_body_parameters = parse.urlencode(create_host_data.get_host_id() | create_host_data.get_annotations())
         params = {}
         params.update(self._default_params)
         return await invoke_endpoint(HttpVerb.POST,
@@ -320,10 +441,14 @@ class Api:
         # get formatted in the url in invoke_endpoint
         params['token'] = token
 
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         return await invoke_endpoint(HttpVerb.DELETE,
                                      ConjurEndpoint.HOST_FACTORY_REVOKE_TOKEN,
                                      params,
-                                     api_token=await self.api_token,
+                                     api_token=api_token,
                                      ssl_verification_metadata=self.ssl_verification_data)
 
     async def set_variable(self, variable_id: str, value: str) -> str:
@@ -336,8 +461,13 @@ class Api:
             'identifier': variable_id,
         }
         params.update(self._default_params)
+
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         response = await invoke_endpoint(HttpVerb.POST, ConjurEndpoint.SECRETS, params,
-                                         value, api_token=await self.api_token,
+                                         value, api_token=api_token,
                                          ssl_verification_metadata=self.ssl_verification_data)
         return response.text
 
@@ -356,8 +486,12 @@ class Api:
         with open(policy_file, 'r') as content_file:
             policy_data = content_file.read()
 
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         response = await invoke_endpoint(http_verb, ConjurEndpoint.POLICIES, params,
-                                         policy_data, api_token=await self.api_token,
+                                         policy_data, api_token=api_token,
                                          ssl_verification_metadata=self.ssl_verification_data)
         return response.json
 
@@ -394,9 +528,14 @@ class Api:
         query_params = {
             'role': resource.full_id()
         }
+
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         response = await invoke_endpoint(HttpVerb.PUT, ConjurEndpoint.ROTATE_API_KEY,
                                          self._default_params,
-                                         api_token=await self.api_token,
+                                         api_token=api_token,
                                          ssl_verification_metadata=self.ssl_verification_data,
                                          query=query_params)
         return response.text
@@ -413,6 +552,26 @@ class Api:
                                          ssl_verification_metadata=self.ssl_verification_data)
         return response.text
 
+    async def set_authenticator_state(self, authenticator_id: str, enabled: bool) -> str:
+        """
+        This method sets the state of a given authenticator
+        """
+
+        params = {
+            'authenticator_id': authenticator_id
+        }
+        params.update(self._default_params)
+
+        body = f'enabled={str(enabled).lower()}'
+
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
+        response = await invoke_endpoint(HttpVerb.PATCH, ConjurEndpoint.AUTHENTICATOR, params, body,
+                                         api_token=api_token, ssl_verification_metadata=self.ssl_verification_data)
+        return response.text
+
     async def change_personal_password(
             self, logged_in_user: str, current_password: str,
             new_password: str) -> str:
@@ -428,6 +587,10 @@ class Api:
         return response.text
 
     async def get_server_info(self):
+        """
+        Provides information about a particular Conjur node
+        @return: Server info in json form
+        """
         params = {
             'url': self._url
         }
@@ -435,13 +598,18 @@ class Api:
                                      ConjurEndpoint.INFO,
                                      params,
                                      ssl_verification_metadata=self.ssl_verification_data)
+
     async def whoami(self) -> dict:
         """
         This method provides dictionary of information about the user making an API request.
         """
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         response = await invoke_endpoint(HttpVerb.GET, ConjurEndpoint.WHOAMI,
                                          self._default_params,
-                                         api_token=await self.api_token,
+                                         api_token=api_token,
                                          ssl_verification_metadata=self.ssl_verification_data)
 
         return response.json
@@ -470,11 +638,15 @@ class Api:
         # Remove 'inspect' from query as it is client-side param that shouldn't get to the server.
         inspect = request_parameters.pop('inspect', None) if request_parameters else None
 
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         response = await invoke_endpoint(HttpVerb.GET,
                                          ConjurEndpoint.ROLES_MEMBERS_OF,
                                          params,
+                                         api_token=api_token,
                                          query=request_parameters,
-                                         api_token=await self.api_token,
                                          ssl_verification_metadata=self.ssl_verification_data)
 
         resources = response.json
@@ -506,10 +678,14 @@ class Api:
         }
         params.update(self._default_params)
 
+        api_token = await self.api_token
+        if api_token is None:
+            raise MissingApiTokenException()
+
         response = await invoke_endpoint(HttpVerb.GET,
                                          ConjurEndpoint.RESOURCES_PERMITTED_ROLES,
                                          params,
-                                         api_token=await self.api_token,
+                                         api_token=api_token,
                                          ssl_verification_metadata=self.ssl_verification_data)
 
         return response.json
