@@ -1,10 +1,21 @@
-from typing import Dict
+"""
+SimpleCredentialsProvider module
+
+This module holds the SimpleCredentialsProvider class
+"""
 from copy import deepcopy
+from typing import Dict
+
 from conjur_api.interface import CredentialsProviderInterface
 from conjur_api.models import CredentialsData
 
 
 class SimpleCredentialsProvider(CredentialsProviderInterface):
+    """
+    SimpleCredentialsProvider class
+
+    This class implement a naive, in-memory implementation of the CredentialsProvider Interface
+    """
 
     def __init__(self):
         self._credentials: Dict[str, CredentialsData] = {}
